package jforex.indicators;

import java.awt.Color;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.TimeZone;

import com.dukascopy.api.IBar;
import com.dukascopy.api.IConsole;
import com.dukascopy.api.IIndicators.MaType;
import com.dukascopy.api.indicators.IDrawingIndicator;
import com.dukascopy.api.indicators.IIndicator;
import com.dukascopy.api.indicators.IIndicatorContext;
import com.dukascopy.api.indicators.IIndicatorDrawingSupport;
import com.dukascopy.api.indicators.IndicatorInfo;
import com.dukascopy.api.indicators.IndicatorResult;
import com.dukascopy.api.indicators.InputParameterInfo;
import com.dukascopy.api.indicators.IntegerRangeDescription;
import com.dukascopy.api.indicators.OptInputParameterInfo;
import com.dukascopy.api.indicators.OutputParameterInfo;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.Stroke;
import java.util.List;
import java.util.Map;
import java.util.Vector;

public class SupDemInd implements IIndicator, IDrawingIndicator {

    private int LOW = 0;
    private int HIGH = 1;
    private IndicatorInfo indicatorInfo;
    // INPUTS
    private InputParameterInfo[] inputParameterInfos;
    private IBar[] inBar;
    // OPT INPUTS 
    private OptInputParameterInfo[] optInputParameterInfos;
    private int period = 7; // period to find high or low
    // OUTPUTS
    private OutputParameterInfo[] outputParameterInfos;
    private Object[][] outDraw;
    // CONTEXT
    private IConsole console;
    
    class PointOnChart {
        public List<Pivot> list;
        public int idx;

        public PointOnChart(List<Pivot> list, int idx) {
            this.list = list;
            this.idx = idx;
        }
    }
    
    class Pivot {
        public IBar bar;
        public int idx;

        public Pivot(IBar bar, int idx) {
            this.bar = bar;
            this.idx = idx;
        }        
    } 

    public int getLookback() {
        return period / 2;
    }

    public int getLookforward() {
        return 0;
    }
    

    public void onStart(IIndicatorContext context) {
        
        this.console = context.getConsole();

        int[] maTypeValues = new int[MaType.values().length];
        String[] maTypeNames = new String[MaType.values().length];
        for (int i = 0; i < maTypeValues.length; i++) {
            maTypeValues[i] = i;
            maTypeNames[i] = MaType.values()[i].name();
        }

        inputParameterInfos = new InputParameterInfo[]{
            new InputParameterInfo("Bar", InputParameterInfo.Type.BAR)
        };

        optInputParameterInfos = new OptInputParameterInfo[]{
            new OptInputParameterInfo("Period",
            OptInputParameterInfo.Type.OTHER,
            new IntegerRangeDescription(period, 3, 100, 2)),};

        outputParameterInfos = new OutputParameterInfo[]{
            new OutputParameterInfo("Low",
            OutputParameterInfo.Type.OBJECT,
            OutputParameterInfo.DrawingStyle.LINE) {
                {
                    setDrawnByIndicator(true);
                }
            },
            new OutputParameterInfo("High",
            OutputParameterInfo.Type.OBJECT,
            OutputParameterInfo.DrawingStyle.LINE) {

                {
                    setDrawnByIndicator(true);
                }
            },
        };

        outDraw = new Object[outputParameterInfos.length][];

        indicatorInfo = new IndicatorInfo("SUP_DEM", "Supply and demand", "My indicators", true, false, false, inputParameterInfos.length, optInputParameterInfos.length, outputParameterInfos.length);
        indicatorInfo.setRecalculateAll(true);
    }

    /**************
     * CALCULATE
     **************/
    public IndicatorResult calculate(int startIdx, int endIdx) {
        
        // startIdx, endIdx - index in input arrays that correspond to the start and end of output

        // output start
        if (startIdx - getLookback() < 0) {
            startIdx = getLookback();
        }
        if (startIdx > endIdx) {
            return new IndicatorResult(0, 0);
        }
        
        // output length
        int len = endIdx - startIdx + 1;
             
        double currentPrice = inBar[inBar.length - 1].getClose();
        List<Pivot> supplyLevels = new Vector<Pivot>();
        List<Pivot> demandLevels = new Vector<Pivot>();
        int calcEnd = endIdx - (period - 1 - getLookback());
        
        
        for (int i = startIdx, outIdx = 0; i <= calcEnd;  i++, outIdx++) {
            
           if(isHighest(inBar, i - getLookback(), period, inBar[i].getHigh())){
                
                double bottomPrice = Math.min(inBar[i].getOpen(), inBar[i].getClose());
                
                while(!supplyLevels.isEmpty()) {
                   int size = supplyLevels.size();
                   // remove support bars that are lower than the current bar
                   if (inBar[i].getHigh() > supplyLevels.get(size - 1).bar.getHigh()) {
                       supplyLevels.remove(size - 1);
                   } else {
                       break;
                   }
               }
                
                if(bottomPrice > currentPrice) {
                    
                    if(supplyLevels.isEmpty()) {
                        // add first bar
                        supplyLevels.add(new Pivot(inBar[i], outIdx));

                    } else {

                        if(supplyLevels.size() > 0) {
                            Pivot pivot = supplyLevels.get(supplyLevels.size() - 1);
                            IBar bar = pivot.bar;
                            if(Math.min(bar.getOpen(), bar.getClose()) > inBar[i].getHigh() ) {
                                supplyLevels.add(new Pivot(inBar[i], outIdx));
                            }

                        } else {
                            supplyLevels.add(new Pivot(inBar[i], outIdx));
                        }
                    }
                
                }
           }
        }        
        
        List<Pivot> result = null;
        int idx = 0;
        
        for(int outIdx = 0; outIdx < len; outIdx++) {
            
            if(idx < supplyLevels.size() && supplyLevels.get(idx).idx <= outIdx) {
                result = supplyLevels.subList(0, idx + 1);
                idx++;
            }
            
            if(result != null) {
                outDraw[LOW][outIdx] = new PointOnChart(result, outIdx);
            } else {
                outDraw[LOW][outIdx] = null;
            }
        }
        
        
        
        for (int i = startIdx, outIdx = 0; i <= calcEnd; i++, outIdx++) {
            
           if(isLowest(inBar, i - getLookback(), period, inBar[i].getLow())){
                
                double topPrice = Math.max(inBar[i].getOpen(), inBar[i].getClose());
                
                while(!demandLevels.isEmpty()) {
                   int size = demandLevels.size();
                   // remove support bars that are lower than the current bar
                   if (inBar[i].getLow() < demandLevels.get(size - 1).bar.getLow()) {
                       demandLevels.remove(size - 1);
                   } else {
                       break;
                   }
               }
                
                if(topPrice < currentPrice) {
                    
                    if(demandLevels.isEmpty()) {
                        // add first bar
                        demandLevels.add(new Pivot(inBar[i], outIdx));

                    } else {
                        if(!demandLevels.isEmpty()) {
                            Pivot pivot = demandLevels.get(demandLevels.size() - 1);
                            IBar bar = pivot.bar;
                            if(Math.max(bar.getOpen(), bar.getClose()) < inBar[i].getLow() ) {
                                demandLevels.add(new Pivot(inBar[i], outIdx));
                            }

                        } else {
                            demandLevels.add(new Pivot(inBar[i], outIdx));
                        }
                    }
                
                }
           }
        } 
        
        result = null;
        idx = 0;
        
        for(int outIdx = 0; outIdx < len; outIdx++) {
            
            if(idx < demandLevels.size() && demandLevels.get(idx).idx <= outIdx) {
                result = demandLevels.subList(0, idx + 1);
                idx++;
            }
            
            if(result != null) {
                outDraw[HIGH][outIdx] = new PointOnChart(result, outIdx);
            } else {
                outDraw[HIGH][outIdx] = null;
            }
        }

        return new IndicatorResult(startIdx, len);
    }

    @Override
    public java.awt.Point drawOutput(
            Graphics g,
            int outputIdx,
            Object values2,
            Color color,
            Stroke stroke,
            IIndicatorDrawingSupport indicatorDrawingSupport,
            List<Shape> shapes,
            Map<Color, List<java.awt.Point>> handles) {
                
        
        Graphics2D g2d = (Graphics2D) g;
        if (values2 != null) {
            
            Object[] values = (Object[]) values2;
            try {
                
                PointOnChart point = (PointOnChart) values[values.length - 1];
                
                int a=0;
                if(point != null) {
                    
                    for(Pivot pivot : point.list) {
                        
                        a = pivot.idx;
                        
                        
                        IBar bar = pivot.bar;
                        
                        double bottom;
                        if(outputIdx == 0) {
                            bottom = Math.min(bar.getOpen(), bar.getClose());
                        } else {
                            bottom = bar.getLow();
                        }
                        
                        double top;
                        if(outputIdx == 0) {           
                            top = bar.getHigh();
                        } else {
                            top = Math.max(bar.getOpen(), bar.getClose());
                        }

                        int nCandles = indicatorDrawingSupport.getNumberOfCandlesOnScreen();
                        int firstCandleIdx = indicatorDrawingSupport.getIndexOfFirstCandleOnScreen();
                        
                        int shift = 0;
                        if(indicatorDrawingSupport.isLastCandleInProgress()) {
                            shift = 1;
                        }
                        
                        int firstIdx = Math.max(firstCandleIdx - shift + nCandles - (point.idx - pivot.idx), firstCandleIdx - 1);
                        
                        int startX = (int) indicatorDrawingSupport.getMiddleOfCandle(firstIdx);
                        int startY = (int) indicatorDrawingSupport.getYForValue(top);
                        
                        int endX = (int) indicatorDrawingSupport.getMiddleOfCandle(firstCandleIdx + nCandles);
                        int endY = (int) indicatorDrawingSupport.getYForValue(bottom);

                        if(outputIdx == 0) {
                            g2d.setColor(Color.red);
                        } else {
                            g2d.setColor(Color.blue);
                        }
                        g2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.3f));
                        g2d.fillRect(startX, startY, endX - startX, endY - startY);

                    }
                }
                
            } catch (Exception e) {
                print(e.getMessage());
                e.printStackTrace();
            }
        }

        return null;
    }

    public boolean isHighest(IBar[] bar, int start, int count, double highValue) {
        for (int i = start; i < start + count; i++) {
            if (Double.compare(bar[i].getHigh(), highValue) > 0) {
                return false;
            }
        }
        return true;
    }

    public boolean isLowest(IBar[] bar, int start, int count, double lowValue) {
        for (int i = start; i < start + count; i++) {
            if (Double.compare(bar[i].getLow(), lowValue) < 0) {
                return false;
            }
        }
        return true;
    }

    public IndicatorInfo getIndicatorInfo() {
        return indicatorInfo;
    }

    public InputParameterInfo getInputParameterInfo(int index) {
        if (index <= inputParameterInfos.length) {
            return inputParameterInfos[index];
        }
        return null;
    }

    public OptInputParameterInfo getOptInputParameterInfo(int index) {
        if (index <= optInputParameterInfos.length) {
            return optInputParameterInfos[index];
        }
        return null;
    }

    public OutputParameterInfo getOutputParameterInfo(int index) {
        if (index <= outputParameterInfos.length) {
            return outputParameterInfos[index];
        }
        return null;
    }

    public void setInputParameter(int index, Object array) {
        inBar = (IBar[]) array;
    }

    public void setOptInputParameter(int index, Object value) {
        switch (index) {
            case 0:
                period = (Integer) value;
                break;
        }
    }

    public void setOutputParameter(int index, Object array) {
        if (array instanceof Object[]) {
            outDraw[index] = (Object[]) array;
        }
    }

    // Some helper methods for printing
    private void print(Object... o) {
        for (Object ob : o) {
            //console.getOut().print(ob + "  ");
            if (ob instanceof double[]) {
                print2((double[]) ob);
            } else if (ob instanceof double[][]) {
                print2((double[][]) ob);
            } else if (ob instanceof Long) {
                print2(toStr((Long) ob));
            } else {
                print2(ob);
            }
            print2(" ");
        }
        console.getOut().println();
    }

    private void print(Object o) {
        console.getOut().println(o);
    }

    private void print2(Object o) {
        console.getOut().print(o);
    }

    private void print(double[] arr) {
        print(arrayToString(arr));
    }

    private void print(double[][] arr) {
        print(arrayToString(arr));
    }

    private void print(Long time) {
        print(toStr(time));
    }

    private void printIndicatorInfos(IIndicator ind) {
        for (int i = 0; i < ind.getIndicatorInfo().getNumberOfInputs(); i++) {
            print(ind.getIndicatorInfo().getName() + " Input " + ind.getInputParameterInfo(i).getName() + " " + ind.getInputParameterInfo(i).getType());
        }
        for (int i = 0; i < ind.getIndicatorInfo().getNumberOfOptionalInputs(); i++) {
            print(ind.getIndicatorInfo().getName() + " Opt Input " + ind.getOptInputParameterInfo(i).getName() + " " + ind.getOptInputParameterInfo(i).getType());
        }
        for (int i = 0; i < ind.getIndicatorInfo().getNumberOfOutputs(); i++) {
            print(ind.getIndicatorInfo().getName() + " Output " + ind.getOutputParameterInfo(i).getName() + " " + ind.getOutputParameterInfo(i).getType());
        }
    }

    public static String arrayToString(double[] arr) {
        String str = "";
        for (int r = 0; r < arr.length; r++) {
            str += "[" + r + "] " + (new DecimalFormat("#.#######")).format(arr[r]) + "; ";
        }
        return str;
    }

    public static String arrayToString(double[][] arr) {
        String str = "";
        if (arr == null) {
            return "null";
        }
        for (int r = 0; r < arr.length; r++) {
            for (int c = 0; c < arr[r].length; c++) {
                str += "[" + r + "][" + c + "] " + (new DecimalFormat("#.#######")).format(arr[r][c]);
            }
            str += "; ";
        }
        return str;
    }

    public String toDecimalToStr(double d) {
        return (new DecimalFormat("#.#######")).format(d);
    }
    
    
    public String toStr(Long time) {
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss") {

            {
                setTimeZone(TimeZone.getTimeZone("GMT"));
            }
        };
        return sdf.format(time);
    }
    
    private String toStr(IBar bar) {
        return toStr(bar.getTime()) + "  O:" + bar.getOpen() + " C:" + bar.getClose() + " H:" + bar.getHigh() + " L:" + bar.getLow();
    }
    
    private void print(IBar bar) {
        print(toStr(bar));
    }
}